from __future__ import annotations
from typing import List
import sys, time, pydevd
from PySide2.QtCore import QObject, QThread, Signal, Slot, QTimer, QEventLoop
from Project.ProjectSettings import ProjectSettings
from Frames.FrameParser import FrameParser
from Frames.Frame import Frame
from Frames.EbdFrame import EbdFrame
from Frames.FrameAddress import FrameAddress
from Utilities.Constants import Constants
from Utilities.Log import Log
from UI.UISignals import UISignals
from Communication.CommandManager import CommandManager
from Communication.SerialPort import SerialPort
import random

class UserApplicationSignals(QObject):
    """This class holds the signals for the user application.
    """
    #: Signal to be emitted when the user application stops
    ApplicationStopped = Signal(bool)

class UserApplication(QThread):
    """Main class for the user application, runs as a thread and manages communication with the FPGA.
    """
    
    TOTAL_BITS = 58176  # Number of faults to inject (10% of overall fault list)

    # Serial port configuration for test results
    __SERIAL_PORT_TEST_REPORT = '/dev/ttyUSB0'  # For Linux (change to COM9 for Windows)
    __SERIAL_PORT_BAUDRATE = 115200  # Baud rate for the serial communication
    __TEST_COMPLETED_TIMEOUT_MSEC = 5000  # Timeout for test completion in milliseconds

    def __init__(self, project: ProjectSettings, frameParser: FrameParser, remoteDeviceSwIdReceived: QObject, parent=None):
        """Initializes the user application with project settings, frame parser, and callback for receiving the SW ID.

        :param project: The project settings object
        :param frameParser: Frame parser for managing frames
        :param remoteDeviceSwIdReceived: Callback for when SW version and ID are received
        :param parent: Optional parent object for the thread
        """
        super().__init__(parent)
        self._execute = False
        self._project = project
        self._timerTestCompleted = None
        self._testResponseError = False
        self._serialPortTestReport = None
        self._eventLoop = None
        self._frameParser = frameParser
        self._commandManager = None
        self._uiSignals = UISignals()
        self._uiSignals.RemoteDeviceSwIdReceived.connect(remoteDeviceSwIdReceived)
        
    def Exit(self):
        """Gracefully exits the user application by stopping the thread and closing the command manager."""
        self.StopClicked()
        try:
            if self._commandManager:
                self._commandManager.Close()  # Close the command manager connection
        except Exception as e:
            Log.PrintException(f'UserApplication.Exit: {str(e)}')  # Log any exceptions during exit

    @Slot(bool)
    def StopClicked(self, value: bool = True):
        """Stops the thread and closes the serial port if open.

        :param value: Optional boolean value passed from the parent object
        """
        if self._eventLoop and self._eventLoop.isRunning():
            self._eventLoop.quit()  # Quit the event loop if it's running
            self._eventLoop = None
        self._execute = False
        if self.isRunning():
            self.terminate()  # Terminate the thread if it's running
        try:
            if self._serialPortTestReport:
                self._serialPortTestReport.Close()  # Close the serial port if it's open
        except Exception as e:
            Log.PrintException(f'UserApplication.StopClicked: {str(e)}')  # Log any exceptions during stop

    def __TimerTestCompletedTimeout(self):
        """Handles the timeout event when a test doesn't complete in time."""
        self._testResponseError = True
        if self._eventLoop:
            self._eventLoop.quit()  # Quit the event loop on timeout

    @Slot(bytes)
    def TestResultReceived(self, dataBytes: bytes):
        """Callback function that processes test results received from the FPGA.

        :param dataBytes: The received bytes from the FPGA
        """
        try:
            value = dataBytes.decode('utf-8')  # Decode the received bytes as a UTF-8 string
            print(f'{value}')
            with open(self._project.Path + "/results.txt", "a+") as f:
                f.write(value)  # Write the test result to the results.txt file
            self._eventLoop.quit()  # Quit the event loop after receiving the result
        except Exception as e:
            Log.PrintException(f'UserApplication.TestResultReceived: {str(e)}')  # Log any exceptions

    def run(self):
        """Main method for running the user application thread. Handles fault injection and test result collection."""
        
        # For debugging with pydevd, uncomment the following lines:
        pydevd.connected = True
        pydevd.settrace(suspend=False)

        # Open the serial port for test results
        self._serialPortTestReport = SerialPort(UserApplication.__SERIAL_PORT_TEST_REPORT, UserApplication.__SERIAL_PORT_BAUDRATE)
        self._serialPortTestReport.Open(self.TestResultReceived)  # Open the serial port and set the callback for received data

        try:
            # Initialize the command manager for communication with the FPGA
            self._commandManager = CommandManager(self._project.FpgaDevice, self._project.IpAddress, self._project.TcpPort)
            self._execute = True

            # Read the board's serial number and software version
            boardInfo = self._commandManager.ReadId()
            serialNumberHex = '{0:0{1}X}'.format(boardInfo.SerialNumber, 8)  # Format serial number as hex
            self._uiSignals.RemoteDeviceSwIdReceived.emit(f'{serialNumberHex} : {boardInfo.Version}')  # Emit SW ID received signal
            
            self.ResetBoard(self)


            # Load the list of frames from the external app_frames.txt file
            app_frames = []
            all_fault_list = []
            with open(self._project.Path + "/app_frames.txt", "r") as f:
                for line in f:
                    tmp = line.strip()  # Remove newlines
                    frm = int(tmp, 16)  # Convert from hex to integer
                    # Generate the fault list (each frame has 3232 bits)
                    for i in range(0, 3232):
                        value = (frm, i)
                        all_fault_list.append(value)
                    app_frames.append(frm)

            # Randomly select frames for fault injection
            faults_inj = random.choices(all_fault_list, k=self.TOTAL_BITS)
            resets=0
            # Inject faults one by one
            injCounter = 0
            for item in faults_inj:
                frameAddress = item[0]
                bitIndex = item[1]

                print(f'Test iteration: {str(injCounter + 1)}')

                # Read the frame from the FPGA
                frames = self._commandManager.ReadFrame(frameAddress, 1)
                # Flip the specific bit in the frame
                frames[0].BitFlip(bitIndex)

                # Inject the fault by writing the frame back to the FPGA
                status = self._commandManager.WriteFrame(frames)
                Log.Print(str(status))  # Log the status

                # Set up a timeout mechanism
                self._testResponseError = False
                self._eventLoop = QEventLoop()
                self._timerTestCompleted = QTimer()
                self._timerTestCompleted.setSingleShot(True)
                self._timerTestCompleted.timeout.connect(self.__TimerTestCompletedTimeout)
                self._timerTestCompleted.start(UserApplication.__TEST_COMPLETED_TIMEOUT_MSEC)

                # Send a command to the FPGA to run the test
                self._serialPortTestReport.Send("X".encode('utf-8'))

                # Wait for the test result or timeout
                self._eventLoop.exec_(QEventLoop.AllEvents)

                # Stop the timer if it's still active
                if self._timerTestCompleted and self._timerTestCompleted.isActive():
                    self._timerTestCompleted.stop()

                # Handle a timeout (test response error)
                if self._testResponseError:
                    self.ResetBoard(self)
                    resets+=1
                    s="Reset at Iteration: "+str(injCounter)+" in Frame/bit Address: "+str(frameAddress)+"/"+str(bitIndex)+"\n"
                    with open(self._project.Path + "/resets.txt", "a+") as reset_f:
                        reset_f.write(s)
                else:
                    # If no timeout, read the frame again and correct the fault
                    frames = self._commandManager.ReadFrame(frameAddress, 1)
                    frames[0].BitFlip(bitIndex)  # Correct the flipped bit
                    status = self._commandManager.WriteFrame(frames)
                    Log.Print(str(status))  # Log the status

                injCounter += 1  # Increment the counter after each fault injection

        except Exception as e:
            Log.PrintException(f'UserApplication.run: {str(e)}')  # Log any exceptions during the run
        self.Exit()  # Exit the application after the test is complete

    @staticmethod
    def ResetBoard(self):
        """Static method to reset and reconfigure the FPGA board."""
        try:
            # Configure the FPGA device using the project's bitstream file
            status = self._commandManager.ConfigureDevice(self._project.BitFilePath)
            Log.Print(str(status))  # Log the configuration status
            
        except Exception as e:
            Log.PrintException(f'UserApplication.ResetBoard: {str(e)}')  # Log any exceptions during the reset
